## Python imports
import logging
from beatbox import PythonClient as SalesforceClient
from beatbox import SessionTimeoutError

## Zope imports
from zope.interface import implements

## Plone imports
from Products.CMFCore.utils import UniqueObject
from OFS.SimpleItem import SimpleItem
from Globals import InitializeClass, DTMLFile
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from AccessControl import ClassSecurityInfo
from Products.CMFCore.permissions import View, ManagePortal

## Interfaces
from interfaces.salesforcebaseconnector import ISalesforceBaseConnector, ISalesforceBaseConnectorInfo

logger = logging.getLogger('SalesforceBaseConnector')

class SalesforceBaseConnector (UniqueObject, SimpleItem):
    """A tool for storing/managing connections and connection information when interacting
       with Salesforce.com via beatbox.
    """
    
    implements(ISalesforceBaseConnector,ISalesforceBaseConnectorInfo)
    
    id = 'portal_salesforcebaseconnector'
    meta_type = 'Salesforce Base Connector'
    title = 'Connect to an external Salesforce instance'

    security = ClassSecurityInfo()

    manage_options=(( { 'label' : 'Configure Authentication'
                        , 'action' : 'manage_config'
                        },
                      ) + SimpleItem.manage_options
                    )
    
    ##   ZMI methods
    security.declareProtected(ManagePortal, 'manage_config')
    manage_config = PageTemplateFile('www/manageAuthConfig', globals() )
    manage_config._owner = None

    ## internal attributes & methods
    _username = ''
    _password = ''
    _v_sfclient = None
    
    def _login(self):
        logger.debug('logging into salesforce...')
        username = self._username
        passwd = self._password
        res = self._v_sfclient.login(username, passwd)
        return res
    
    def _areValidCredentials(self, username, passwd):
        logger.debug('testing new user credentials...')
        try:
            testClient = SalesforceClient()
            testClient.login(username,passwd)
            return True
        except:
            return False

    def _getClient(self):
        logger.debug('calling _getClient')
        if self._v_sfclient is None:
            self._v_sfclient = SalesforceClient()
        if not self._v_sfclient.isConnected():
            logger.debug('No open connection to Salesforce. Trying to log in...')
            response = self._login()
            if not response:
                raise "Salesforce login failed"
        return self._v_sfclient 

    def _resetClient(self):
        logger.debug('reseting client')
        self._v_sfclient = None
        
        
    security.declareProtected(ManagePortal, 'manage_configSalesforceCredentials')
    def manage_configSalesforceCredentials(self, username, password, REQUEST=None):
        """Called by the ZMI auth management tab """
        portalMessage = ''
        if self.setCredentials(username, password):
            portalMessage = 'Authentication tested successfully. Username and password saved.'
        else:
            portalMessage = 'The supplied credentials could not be authenticated. No credentials saved.'
        if REQUEST is not None:
            REQUEST.RESPONSE.redirect('%s/manage_config?portal_status_message=%s' % (self.absolute_url(),portalMessage))
        
    security.declareProtected(ManagePortal, 'setCredentials')
    def setCredentials(self, username, password):
        if self._areValidCredentials(username, password):
            self._username = username
            self._password = password
            # Disconnect from any previously connected Salesforce instance
            self._resetClient()            
            return True
        return False

    security.declareProtected(ManagePortal, 'setBatchSize')
    def setBatchSize(self, batchsize):
        """Set the batchsize used by query and queryMore"""
        try:
            self._getClient().batchSize = batchsize
        except SessionTimeoutError:
            self._resetClient()
            self._getClient().batchSize = batchsize
    
    security.declareProtected(ManagePortal, 'getUsername')    
    def getUsername(self):
        """Return the current stored Salesforce username"""
        return self._username

    security.declareProtected(ManagePortal, 'getPassword')    
    def getPassword(self):
        """Return the current stored Salesforce password"""
        return self._password
    
    ##
    # Salesforce API
    ##
    
    ## Accessors
    security.declarePublic('query')
    def query(self, fieldList, sObjectType, whereClause=''):
        """Run a SoQL query against a Salesforce instance:
           portal.portal_salesforcebaseconnector.query(
                    ['FirstName',],'Contact',"FirstName='Stella'")
           """
        logger.debug('calling query()')
        if sObjectType is None:
            raise "Invalid argument: sObjectType must not be None"
        if fieldList:
            fieldString = ','.join(fieldList)
        try:
            result = self._getClient().query(fieldString, sObjectType, whereClause)
        except SessionTimeoutError:
            self._resetClient()
            result = self._getClient().query(fieldString, sObjectType, whereClause)
            
        return result
    
    security.declareProtected(ManagePortal, 'describeGlobal')    
    def describeGlobal(self):
        '''Retrieves list of available objects'''
        logger.debug('calling describeGlobal')
        try:
            result = self._getClient().describeGlobal()
        except SessionTimeoutError:
            self._resetClient()
            result = self._getClient().describeGlobal()
        
        return result
        
    security.declareProtected(ManagePortal, 'describeSObjects')            
    def describeSObjects(self, sObjectTypes):
        '''Introspection on Salesforce datatypes
        Currently this method returns a list of beatbox SObjects:
        
        [<beatbox.python_client.SObject object at 0x04F515F0>]
        TODO: should we pull out the contents and return that?
        '''
        logger.debug('calling describeSObjects')
        try:
            result = self._getClient().describeSObjects(sObjectTypes)
        except SessionTimeoutError:
            self._resetClient()
            result = self._getClient().describeSObjects(sObjectTypes)
        
        return result        
        
    security.declarePublic('queryMore')        
    def queryMore(self, queryLocator):
        '''Query more record(s) in Salesforce'''
        logger.debug('calling queryMore')
        try:
            result = self._getClient().queryMore(queryLocator)
        except SessionTimeoutError:
            self._resetClient()
            result = self._getClient().queryMore(queryLocator)
        
        return result
        
    security.declarePublic('retrieve')        
    def retrieve(self, fields, sObjectType, ids):
        '''Retrieve record(s) in Salesforce based on one or more SF Object ID's
        t.retrieve(['FirstName','Id'],'Contact','0033000000JKDFxAAP')
            or
        t.retrieve(['FirstName','Id'],'Contact',['0033000000JKDFxAAP'])         
        returns: [{'Id': '0033000000JKDFxAAP', 'FirstName': 'Siddartha'}] 
        '''
        logger.debug('calling retrieve')
        fieldString = ''
        if fields:
            fieldString = ','.join(fields)
        try:
            result = self._getClient().retrieve(fieldString, sObjectType, ids)
        except SessionTimeoutError:
            self._resetClient()
            result = self._getClient().retrieve(fieldString, sObjectType, ids)
        
        return result        
        
    security.declarePublic('getDeleted')        
    def getDeleted(self, sObjectType, start, end):
        '''Get Deleted Salesforce Objects'''
        logger.debug('calling getDeleted')
        try:
            result = self._getClient().getDeleted(sObjectType, start, end)
        except SessionTimeoutError:
            self._resetClient()
            result = self._getClient().getDeleted(sObjectType, start, end)
        
        return result
    
    security.declarePublic('getUpdated')        
    def getUpdated(self, sObjectType, start, end):
        '''Get Recently Updated Salesforce Objects'''
        logger.debug('calling getUpdated')
        try:
            result = self._getClient().getUpdated(sObjectType, start, end)
        except SessionTimeoutError:
            self._resetClient()
            result = self._getClient().getUpdated(sObjectType, start, end)
        
        return result
    
    security.declareProtected(ManagePortal, 'getUserInfo')    
    def getUserInfo(self):
        '''Retrieves information about the authenticated user and
        returns it as a dictionary'''
        logger.debug('calling getUserInfo')
        try:
            result = self._getClient().getUserInfo()
        except SessionTimeoutError:
            self._resetClient()
            result = self._getClient().getUserInfo()
        
        return result

    security.declareProtected(ManagePortal, 'describeTabs')        
    def describeTabs(self):
        '''Describe the apps that have been configured for the user'''
        logger.debug('calling describeTabs')
        try:
            result = self._getClient().describeTabs()
        except SessionTimeoutError:
            self._resetClient()
            result = self._getClient().describeTabs()
        
        return result
           

    ## Mutators
    security.declareProtected(ManagePortal, 'create')    
    def create(self, sObjects):
        '''Create record(s) in Salesforce
        sObjects are one or more dictionaries like this:
        obj = dict(type='Contact',
            LastName='Doe',
            FirstName='John',
            Phone='123-456-7890',
            Email='john@doe.com',
            Birthdate = datetime.date(1970, 1, 4)
        )'''
        logger.debug('calling create')
        try:
            result = self._getClient().create(sObjects)
        except SessionTimeoutError:
            self._resetClient()
            result = self._getClient().create(sObjects)
        
        return result
        
    security.declareProtected(ManagePortal, 'update')            
    def update(self, sObjects):
        '''Update record(s) in Salesforce
        data = dict(type='Contact',
            Id=id,
            Birthdate = newdate)
            '''
        logger.debug('calling update')
        try:
            result = self._getClient().update(sObjects)
        except SessionTimeoutError:
            self._resetClient()
            result = self._getClient().update(sObjects)
        
        return result

    security.declareProtected(ManagePortal, 'upsert')    
    def upsert(self, externalIdName, sObjects):
        '''If record(s) already exist in Salesforce an update is performed.
        Otherwise, new record(s) are created.'''
        logger.debug('calling upsert')
        try:
            result = self._getClient().upsert(externalIdName, sObjects)
        except SessionTimeoutError:
            self._resetClient()
            result = self._getClient().upsert(externalIdName, sObjects)
        
        return result

    security.declareProtected(ManagePortal, 'delete')    
    def delete(self, ids):
        '''Delete record(s) in Salesforce'''        
        logger.debug('calling delete')
        try:
            result = self._getClient().delete(ids)
        except SessionTimeoutError:
            self._resetClient()
            result = self._getClient().delete(ids)
        
        return result
    
InitializeClass(SalesforceBaseConnector)